package minimizeDipole;

import java.io.*;
import java.util.ArrayList;
import java.lang.StringBuffer;

import static minimizeDipole.Driver.*;


public class Crystal {
	
	ArrayList<Oxygen> oxyList = new ArrayList<>();
	ArrayList<Hydrogen> allHydList = new ArrayList<>();
	ArrayList<Hydrogen> hydList = new ArrayList<>();
	private BufferedReader file = null;
	private BufferedWriter PDBWriter = null;
	private ArrayList<String> lines = new ArrayList<>();
	double[] dipole = new double[]{0,0,0};
	
	int firstFew = 4000;
	
	
	Crystal(String filePath) throws IOException {
		readFile(filePath);
		parseFile();
	}
	
	private void readFile(String filePath) throws IOException {
		file = new BufferedReader(new FileReader(filePath));
		String line;
		while ( (line = file.readLine()) != null) {
			lines.add(line);
		}
		file.close();
	}
	
	private void parseFile() {
		String line, type;
		double x, y, z;
		for ( int i = 0 ; i < lines.size() ; i++ ) {
			line = lines.get(i);
			if ( line.indexOf(" ") != -1 ) {
				type = line.substring(0,line.indexOf(" "));
			} else {
				type = line;
			}
			if ( type.equals("ATOM") || type.equals("HETATM") ) {
				for( int j = 0 ; j < 6 ; j++ ) {
					line = line.substring(line.indexOf(" ")).trim();
				}
				x = Double.parseDouble(line.substring(0,line.indexOf(" ")));
				line = line.substring(line.indexOf(" ")).trim();
				y = Double.parseDouble(line.substring(0,line.indexOf(" ")));
				line = line.substring(line.indexOf(" ")).trim();
				z = Double.parseDouble(line.substring(0,line.indexOf(" ")));
				oxyList.add(new Oxygen(x,y,z));
			}
			
		}
	}

	public void findOxyNeighbors() {
		for ( int i = 0 ; i < oxyList.size() ; i++ ) {
			Hydrogen[] newHydrogens = oxyList.get(i).findNeighbors(oxyList);
			for ( int j = 0 ; j < 4 ; j++ ) {
				allHydList.add(newHydrogens[j]);
			}
		}
	}

	public String toString() {
		StringBuffer output = new StringBuffer();
		StringBuffer current = new StringBuffer();
		String ID, x, y, z;
		int IDSize, xSize, ySize, zSize;
		double[] coord;
		int atomNumber = 1;
		double box = 11.877;
		int A = 2;
		int B = 2;
		int C = 1;
		int copyNumber = 0;
		output.append("COMPND    UNNAMED\n");
		output.append("AUTHOR    SpiderPig\n");
		output.append("CRYST1   "+2*CENTER[0]+"   "+2*CENTER[1]+"   "+2*CENTER[2]+"  90.00  90.00  90.00 P1          \n");
		for (int a = 0 ; a < A ; a++) {
			for (int b = 0 ; b < B ; b++) {
				for (int c = 0 ; c < C ; c++) {
					for ( int i = 0 ; i < oxyList.size() ; i++ ) {
						current.append("                                       ");
						current.append("                                       ");
						current.insert(0, "ATOM   ");
						ID = String.valueOf(atomNumber++);
						IDSize = ID.length();
						current.replace(11-IDSize, 11, ID);
						current.replace(13, 25, "OH2 TP4E    0");
						ID = String.valueOf(i+1+(oxyList.size()*copyNumber));
						IDSize = ID.length();
						current.replace(26-IDSize, 26, ID);
						coord = oxyList.get(i).coord;
						x = String.format("%1$.3f", coord[0]+box*a);
						y = String.format("%1$.3f", coord[1]+box*b);
						z = String.format("%1$.3f", coord[2]+box*c);
						xSize = x.length();
						ySize = y.length();
						zSize = z.length();
						current.replace(38-xSize, 38, x);
						current.replace(46-ySize, 46, y);
						current.replace(54-zSize, 54, z);
						current.insert(56, "1.00  0.00      HYD");
						current.append("\n");
						output.append(current);
						current.setLength(0);
						for (int j = 0 ; j < 2 ; j++) {
							current.append("                                       ");
							current.append("                                       ");
							current.insert(0, "ATOM   ");
							ID = String.valueOf(atomNumber++);
							IDSize = ID.length();
							current.replace(11-IDSize, 11, ID);
							current.replace(13, 25, "H"+ (j+1) +"  TP4E    0");
							ID = String.valueOf(i+1+(oxyList.size()*copyNumber));
							IDSize = ID.length();
							current.replace(26-IDSize, 26, ID);
							coord = oxyList.get(i).myHydrogens[j].coord;
							x = String.format("%1$.3f", coord[0]+box*a);
							y = String.format("%1$.3f", coord[1]+box*b);
							z = String.format("%1$.3f", coord[2]+box*c);
							xSize = x.length();
							ySize = y.length();
							zSize = z.length();
							current.replace(38-xSize, 38, x);
							current.replace(46-ySize, 46, y);
							current.replace(54-zSize, 54, z);
							current.insert(56, "1.00  0.00      HYD");
							current.append("\n");
							output.append(current);
							current.setLength(0);
						}
						current.append("                                       ");
						current.append("                                       ");
						current.insert(0, "ATOM   ");
						ID = String.valueOf(atomNumber++);
						IDSize = ID.length();
						current.replace(11-IDSize, 11, ID);
						current.replace(13, 25, "OM  TP4E    0");
						ID = String.valueOf(i+1+(oxyList.size()*copyNumber));
						IDSize = ID.length();
						current.replace(26-IDSize, 26, ID);
						coord = oxyList.get(i).lonePair;
						x = String.format("%1$.3f", coord[0]+box*a);
						y = String.format("%1$.3f", coord[1]+box*b);
						z = String.format("%1$.3f", coord[2]+box*c);
						xSize = x.length();
						ySize = y.length();
						zSize = z.length();
						current.replace(38-xSize, 38, x);
						current.replace(46-ySize, 46, y);
						current.replace(54-zSize, 54, z);
						current.insert(56, "1.00  0.00      HYD");
						current.append("\n");
						output.append(current);
						current.setLength(0);
					}
					copyNumber++;
				}
			}
		}
		
				output.append("MASTER        0    0    0    0    0    0    0    0  544    0  544    0 \n");
		output.append("END \n");
		return output.toString();
	}

	public void toPDB(String outputPath) throws IOException {
		PDBWriter = new BufferedWriter(new FileWriter(outputPath));
		PDBWriter.write(toString());
		PDBWriter.close();
	}

	public void minimizeDipole() {
		Oxygen oxy;
		updateHistory(-1);
		for ( int i = 0 ; i < oxyList.size() ; i++ ) {
			if (i > firstFew-1) {
				break;
			}
//			System.out.println(oxyList.size());
			System.out.println(i);
			oxy = oxyList.get(i);
			dipole = oxy.minimizeDipole(dipole);
			if ( dipole != null ) {
				updateHistory(i);
//				System.out.println(i + " Good");
			} else {
				int option = oxy.bestOption;
//				System.out.println(i + " Bad");
				while (option >= oxy.keys.size()-1) {
					oxy.bestOption = 0;
					i=i-1;
//					System.out.println(i + " Bad");
					oxy = oxyList.get(i);
					option = oxy.bestOption;
				}
//				System.out.println("Here in option " + oxy.keys.size());
				oxy.bestOption++;
				backtrack(i);
				dipole = oxy.oldDipole;
				i--;

			}
			System.out.println(length(dipole));
		}
	}
	
	private void updateHistory(int i) {
		Oxygen oxy;
		for ( int j = 0 ; j < oxyList.size() ; j++ ) {
			oxy = oxyList.get(j);
			oxy.obligationHistory[i+1] = oxy.obligation;
			oxy.curPartnerHistory[i+1] = oxy.curPartner;
		}
	}

//	private boolean checkIceRules(int index) {
//		boolean flag = true;
//		Oxygen oxy;
//		for (int i = 0 ; i < index+1 ; i++ ) {
//			oxy = oxyList.get(i);
//
//			for ( Oxygen neighbor : oxy.neighbors) {
//				if ( neighbor.myHydrogens[0] != null) {
//					if ( Arrays.asList(neighbor.partners).contains(oxy) 
//							&& Arrays.asList(oxy.partners).contains(neighbor) ) {
//						flag = false;
//						break;
//					}
//					if ( Arrays.asList(neighbor.pointers).contains(oxy) 
//							&& Arrays.asList(oxy.pointers).contains(neighbor) ) {
//						flag = false;
//						break;
//					}
//				} 
//			}
//			if ( oxy.obligation > 2 || oxy.curPartner > 2 ) {
//				flag = false;
//				break;
//			}
//			if ( i > index && ( oxy.obligation == 2 || oxy.curPartner == 2 ) ) {
//				oxyList.remove(i);
//				oxyList.add(index+1,oxy);
//			}
//		}
//		return flag;
//	}

	private void backtrack(int i) {
		Oxygen oxy;
		for ( int j = 0 ; j < oxyList.size() ; j++ ) {
			oxy = oxyList.get(j);
			oxy.obligation = oxy.obligationHistory[i];
			oxy.curPartner = oxy.curPartnerHistory[i];
		}
		for ( int k = i+1 ; k < oxyList.size() ; k++ ) {
			oxy = oxyList.get(k);
			oxy.partners = new Oxygen[2];
			oxy.pointers = new Oxygen[2];
			oxy.myHydrogens = new Hydrogen[2];
			oxy.bestOption = 0;
			oxy.keys.clear();
			oxy.dipoles.clear();
			oxy.oldDipole = new double[3];
		}
		oxy = oxyList.get(i);
		oxy.partners = new Oxygen[2];
		oxy.pointers = new Oxygen[2];
	}	
	
//	public void minimizeDipole() {
//		for ( int i = 0 ; i < oxyList.size() ; i++ ) {
//			System.out.println("i = " + i);
//			if (i > firstFew-1) {
//				break;
//			}
//			dipole = oxyList.get(i).minimizeDipole(dipole);
//			setInStone();
//			int changes = setInStone();
//			while ( changes > 0 ) {
//				changes = setInStone();
//			}
//			System.out.println(length(dipole));
//			for ( int k = 0 ; k < oxyList.size() ; k++ ) {
//				if (oxyList.get(k).curPartner == 2 || 
//						oxyList.get(k).obligation == 2) {
//					dipole = oxyList.get(k).minimizeDipole(dipole);
//				}
//			}
//		}
//	}
	
//	public void minimizeDipole() {
//		Oxygen oxy;
//		int fullObligation, fullCurPartner;
//		HashSet<Oxygen> toProcess = new HashSet<Oxygen>();
//		for ( int k = 0 ; k < oxyList.size() ; k++ ) {
//			toProcess.add(oxyList.get(k));
//			System.out.println("k = " + k);
//			if (k > firstFew-1) {
//				break;
//			}
//			while ( ! toProcess.isEmpty() ) {
//				for ( Oxygen oxygen : toProcess ) {
//					dipole = oxygen.minimizeDipole(dipole);
//				}
//				toProcess.clear();
//				for ( int j = 0 ; j < oxyList.size() ; j++ ) {
//					oxy = oxyList.get(j);
//					if ( oxy.curPartner == 2 || oxy.obligation == 2) {
//						toProcess.add(oxy);
//					} else {
//						fullObligation = 0;
//						fullCurPartner = 0;
//						for ( Oxygen neighbor : oxy.neighbors ) {
//							if (neighbor.obligation == 2 ) {
//								fullObligation++;
//							} 
//							if ( neighbor.curPartner == 2 ) {
//								fullCurPartner++;
//							}
//						}
//						if ( oxy.obligation == 1 && fullObligation == 1 ) {
//							toProcess.add(oxy);
//						} else if ( oxy.curPartner == 1 && fullCurPartner == 1 ) {
//							toProcess.add(oxy);
//						} else if ( fullObligation == 2 ) {
//							toProcess.add(oxy);
//						} else if ( fullCurPartner == 2 ) {
//							toProcess.add(oxy);
//						}
//					}
//				}
//			}
//		}
//	}
	
		//				if ( oxy.curPartner == 2 || oxy.obligation == 2 || flag1 || flag2 ) {
//					dipole = oxyList.get(k).minimizeDipole(dipole);
//					setInStone();
//					if (dipole != oldDipole) {
//						changes++;
//						oldDipole = dipole;
//					} 
//				}
//			}
//		}
//		return changes;	
	
	public void fixAllAngles() {
		for (Oxygen oxy : oxyList) {
			oxy.fixAngles();
		}
	}

	private static double length(double[] dipole) {
		double x = dipole[0];
		double y = dipole[1];
		double z = dipole[2];
		return Math.sqrt((x*x)+(y*y)+(z*z));
	}
	
	public void checkNeighborCount() {
		System.out.println("Start checking Neigbhor Count");
		for ( int i = 0 ; i < oxyList.size() ; i++ ) {
			if ( oxyList.get(i).count < 4 ) {
				System.out.println(oxyList.get(i).count);
			}
		}
		System.out.println("End checking Neigbhor Count");
	}
	

	
	
	
	
}
