package minimizeDipole;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;

import static minimizeDipole.Driver.*;

public class Oxygen implements Comparable<Oxygen>{

	double[] coord = new double[3];
	double[] HPos1 = new double[3];
	double[] HPos2 = new double[3];
	Oxygen[] neighbors = new Oxygen[4];
	int count = 0;
	boolean flag = false;
	double[] lonePair = new double[3];
	
	Oxygen[] partners = new Oxygen[2];
	Oxygen[] pointers = new Oxygen[2];
	int[] curPartnerHistory = new int[600];
	int[] obligationHistory = new int[600];
	int curPartner = 0;
	int obligation = 0;
	int bestOption = 0;
	ArrayList<Double> keys = new ArrayList<>();
	HashMap<Double,Double[]> dipoles =  new HashMap<>();
	Hydrogen[] allHydrogens = new Hydrogen[4];
	Hydrogen[] myHydrogens = new Hydrogen[2];
	
	double[] oldDipole = null;
	
	
	Oxygen(double x, double y, double z) {
		coord[0] = x;
		coord[1] = y;
		coord[2] = z;
	}
	
	public String toString() {
		String output = "";
		for ( int i = 0 ; i < 3 ; i++ ) {
			output += coord[i] + "  ";
		}
		return output;
	}
	
	public Hydrogen[] findMyNeighbors(ArrayList<Oxygen> oxyList) {
		double distance;
		Oxygen current;
		Hydrogen[] myBabies = new Hydrogen[4];
		for ( int i = 0 ; i < oxyList.size() ; i++ ) {
			current = oxyList.get(i);
			distance = 0;
			distance += Math.pow( coord[0] - current.coord[0] , 2);
			distance += Math.pow( coord[1] - current.coord[1] , 2);
			distance += Math.pow( coord[2] - current.coord[2] , 2);
			distance = Math.pow( distance , 0.5 );
			if ( distance > 0 && distance < OXY_SPACING ) {
				System.out.println(current);
				neighbors[count] = current;
				myBabies[count] = insertHydrogen(current);
				count++;
			}
		}
		return myBabies;
	}
	
	public int compareTo(Oxygen otherOxygen) {
		return (int)(1000*(distFromCenter() - otherOxygen.distFromCenter()));
	}
	
	private double distFromCenter() {
		double distance = 0.0;
		distance += Math.pow( coord[0] - CENTER[0] , 2);
		distance += Math.pow( coord[1] - CENTER[1] , 2);
		distance += Math.pow( coord[2] - CENTER[2] , 2);
//		distance += Math.pow( coord[0] , 2);
//		distance += Math.pow( coord[1] , 2);
//		distance += Math.pow( coord[2] , 2);
		distance = Math.pow( distance , 0.5 );
		return distance;
	}

	public Hydrogen insertHydrogen(Oxygen neighbor) {
		double distFromNeighbor, x, y, z;
		distFromNeighbor = distanceFrom(neighbor);
		x = coord[0] + ( neighbor.coord[0] - coord[0] ) * (BOND_LENGTH / distFromNeighbor);
		y = coord[1] + ( neighbor.coord[1] - coord[1] ) * (BOND_LENGTH / distFromNeighbor);
		z = coord[2] + ( neighbor.coord[2] - coord[2] ) * (BOND_LENGTH / distFromNeighbor);
		return new Hydrogen(x,y,z);
	}
	
	private double distanceFrom(Oxygen neighbor) {
		double distance = 0;
		distance += Math.pow( coord[0] - neighbor.coord[0] , 2);
		distance += Math.pow( coord[1] - neighbor.coord[1] , 2);
		distance += Math.pow( coord[2] - neighbor.coord[2] , 2);
		distance = Math.pow( distance , 0.5 );
		return distance;
	}

	public Hydrogen[] findNeighbors(ArrayList<Oxygen> oxyList) {
		allHydrogens = findMyNeighbors(oxyList);
		if ( count < 4 ) {
			Hydrogen[] myBabies = new Hydrogen[4];
			for ( int i = 0 ; i < 3 ; i++ ) {
				if ( coord[i] < CENTER[i] ) {
					coord[i] += 2*CENTER[i];
					myBabies = findMyNeighbors(oxyList);
					coord[i] -= 2*CENTER[i];
					for ( int j = 0 ; j < 4 ; j++ ) {
						Hydrogen baby = myBabies[j];
						if ( baby != null ) {
							baby.coord[i] -= 2*CENTER[i];
							allHydrogens[j] = baby;
						}
					}
				} else {
					coord[i] -= 2*CENTER[i];
					myBabies = findMyNeighbors(oxyList);
					coord[i] += 2*CENTER[i];
					for ( int j = 0 ; j < 4 ; j++ ) {
						Hydrogen baby = myBabies[j];
						if ( baby != null ) {
							baby.coord[i] += 2*CENTER[i];
							allHydrogens[j] = baby;
						}
					}
				}
			}
		}
		return allHydrogens;
	}

	public double[] minimizeDipole(double[] dipole) {
		oldDipole = dipole;
		findOptions(dipole);
		if (keys.size() == 0) {
			return null;
		}
		Double[] minDipole = new Double[3];
		minDipole = dipoles.get(keys.get(bestOption));
		myHydrogens[0] = allHydrogens[(int)(double)minDipole[0]];
		myHydrogens[1] = allHydrogens[(int)(double)minDipole[1]];
		partners[0] = neighbors[(int)(double)minDipole[0]];
		partners[1] = neighbors[(int)(double)minDipole[1]];
		double[] newDipole = new double[] {(double)minDipole[3],(double)minDipole[4],(double)minDipole[5]};
		int pointerCount = 0;
		for ( int i = 0 ; i < 4 ; i++ ) {
			if (neighbors[i] != partners[0] && neighbors[i] != partners[1]) {
				neighbors[i].obligation++;
				pointers[pointerCount++] = neighbors[i];
			} else {
				neighbors[i].curPartner++;
			}
		}
//		obligation = 2;
//		curPartner = 2;
		return newDipole;
	}
	
	public void fixAngles() {
		double[] midpoint = new double[3];
		double[] h1 = myHydrogens[0].coord;
		double[] h2 = myHydrogens[1].coord;
		midpoint[0] = (h1[0]+h2[0])/2;
		midpoint[1] = (h1[1]+h2[1])/2;
		midpoint[2] = (h1[2]+h2[2])/2;
		double a = 0;
		a += Math.pow(midpoint[0]-coord[0], 2);
		a += Math.pow(midpoint[1]-coord[1], 2);
		a += Math.pow(midpoint[2]-coord[2], 2);
		a = Math.pow(a,0.5);
		lonePair[0] = coord[0] + (midpoint[0]-coord[0])/a*LONE_PAIR_RADIUS;
		lonePair[1] = coord[1] + (midpoint[1]-coord[1])/a*LONE_PAIR_RADIUS;
		lonePair[2] = coord[2] + (midpoint[2]-coord[2])/a*LONE_PAIR_RADIUS;
		double b = a*Math.tan(Math.toRadians(BOND_ANGLE/2));
		double c = 0;
		c += Math.pow(midpoint[0]-h1[0], 2);
		c += Math.pow(midpoint[1]-h1[1], 2);
		c += Math.pow(midpoint[2]-h1[2], 2);
		c = Math.pow(c,0.5);
		double[] h1p = new double[3];
		double[] h2p = new double[3];
		h1p[0] = midpoint[0] + (midpoint[0]-h1[0])*b/c;
		h1p[1] = midpoint[1] + (midpoint[1]-h1[1])*b/c;
		h1p[2] = midpoint[2] + (midpoint[2]-h1[2])*b/c;
		h2p[0] = midpoint[0] + (midpoint[0]-h2[0])*b/c;
		h2p[1] = midpoint[1] + (midpoint[1]-h2[1])*b/c;
		h2p[2] = midpoint[2] + (midpoint[2]-h2[2])*b/c;
		double d = 0;
		d += Math.pow(h1p[0]-coord[0], 2);
		d += Math.pow(h1p[1]-coord[1], 2);
		d += Math.pow(h1p[2]-coord[2], 2);
		d = Math.pow(d,0.5);
		double[] h1f = new double[3];
		double[] h2f = new double[3];
		h1f[0] = coord[0] + (h1p[0]-coord[0])/d*BOND_LENGTH;
		h1f[1] = coord[1] + (h1p[1]-coord[1])/d*BOND_LENGTH;
		h1f[2] = coord[2] + (h1p[2]-coord[2])/d*BOND_LENGTH;
		h2f[0] = coord[0] + (h2p[0]-coord[0])/d*BOND_LENGTH;
		h2f[1] = coord[1] + (h2p[1]-coord[1])/d*BOND_LENGTH;
		h2f[2] = coord[2] + (h2p[2]-coord[2])/d*BOND_LENGTH;
		myHydrogens[0] = new Hydrogen(h1f);
		myHydrogens[1] = new Hydrogen(h2f);
	}
	
//	private void findOptions(double[] dipole) {
//		if ( bestOption != 0) {
//			return;
//		}
//		double a, b, c, x, y, z, m;
//		for ( int i = 0 ; i < 4 ; i++ ) {
//			if (allHydrogens[i] == null) {
//				continue;
//			}
//			for ( int j = i+1 ; j < 4 ; j++ ) {
//				if(allHydrogens[j] == null) {
//					continue;
//				}
//				// center around oxy
//				a = allHydrogens[i].coord[0] - coord[0];
//				b = allHydrogens[i].coord[1] - coord[1];
//				c = allHydrogens[i].coord[2] - coord[2];
//				x = allHydrogens[j].coord[0] - coord[0];
//				y = allHydrogens[j].coord[0] - coord[1];
//				z = allHydrogens[j].coord[0] - coord[2];
//				x = -x-a;
//				y = -y-b;
//				z = -z-c;
//				m = Math.sqrt((x*x)+(y*y)+(z*z));
//				x = (x/m*MOL_DIPOLE) + dipole[0];
//				y = (y/m*MOL_DIPOLE) + dipole[1];
//				z = (z/m*MOL_DIPOLE) + dipole[2];
//				m = Math.sqrt((x*x)+(y*y)+(z*z));
//				if ( dipoles.containsKey(m) ) {
//					m = m*(1.000000001+0.000000001*Math.random());
//				}
//				dipoles.put(m, new Double[] {(double) i, (double) j,m,x,y,z});
//			}
//		}
//		keys.addAll(dipoles.keySet());
//		Collections.sort(keys);
//		
//	}

//	private void findOptions(double[] dipole) {
//		if ( bestOption != 0) {
//			return;
//		}
//		for ( int i = 0 ; i < 4 ; i++ ) {
//			// if my neighbor has been processed
//			if (neighbors[i].myHydrogens[0] != null) {
//				Oxygen oxy1 = neighbors[i].partners[0];
//				Oxygen oxy2 = neighbors[i].partners[1];
//				// if it is pointing at me
//				if (this == oxy1 || this == oxy2) {
//					allHydrogens[i] = null;
//				} 
//			}
//		}
//		for ( int i = 0 ; i < 4 ; i++ ) {
//			// if neighbor is not processed
//			if (neighbors[i].myHydrogens[0] == null) {
//				if ( neighbors[i].curPartner == 2 ) {
//					allHydrogens[i] = null;
//				} 
//			}
//		}
//
//		double a, b, c, x, y, z, m;
//		for ( int i = 0 ; i < 4 ; i++ ) {
//			if (allHydrogens[i] == null) {
//				continue;
//			}
//			for ( int j = i+1 ; j < 4 ; j++ ) {
//				if(allHydrogens[j] == null) {
//					continue;
//				}
//				// center around oxy
//				a = allHydrogens[i].coord[0] - coord[0];
//				b = allHydrogens[i].coord[1] - coord[1];
//				c = allHydrogens[i].coord[2] - coord[2];
//				x = allHydrogens[j].coord[0] - coord[0];
//				y = allHydrogens[j].coord[0] - coord[1];
//				z = allHydrogens[j].coord[0] - coord[2];
//				x = -x-a;
//				y = -y-b;
//				z = -z-c;
//				m = Math.sqrt((x*x)+(y*y)+(z*z));
//				x = (x/m*MOL_DIPOLE) + dipole[0];
//				y = (y/m*MOL_DIPOLE) + dipole[1];
//				z = (z/m*MOL_DIPOLE) + dipole[2];
//				m = Math.sqrt((x*x)+(y*y)+(z*z));
//				if ( dipoles.containsKey(m) ) {
//					m = m*(1.000000001+0.000000001*Math.random());
//				}
//				dipoles.put(m, new Double[] {(double) i, (double) j,m,x,y,z});
//			}
//		}
//		keys.addAll(dipoles.keySet());
//		Collections.sort(keys);
//		
//	}
	
	private void findOptions(double[] dipole) {
		if ( bestOption != 0) {
			return;
		}
		int index = 0;
		int[] myHydIndex = new int[2];
		Hydrogen[] allHydrogens1 = new Hydrogen[] 
				{allHydrogens[0],allHydrogens[1],allHydrogens[2],allHydrogens[3]};
		for ( int i = 0 ; i < 4 ; i++ ) {
			// if my neighbor has been processed
			if (neighbors[i].myHydrogens[0] != null) {
				Oxygen oxy1 = neighbors[i].partners[0];
				Oxygen oxy2 = neighbors[i].partners[1];
				// if it is pointing at me
				if (this == oxy1 || this == oxy2) {
					allHydrogens1[i] = null;
				} else {
					if (index == 2) {
						return;
					}
					myHydrogens[index] = allHydrogens1[i];
					myHydIndex[index++] = i;
					allHydrogens1[i] = null;
				}
			}
		}
		for ( int i = 0 ; i < 4 ; i++ ) {
			// if neighbor is not processed
			if (neighbors[i].myHydrogens[0] == null) {
				if ( neighbors[i].obligation == 2 ) {
					if (index == 2) {
						return;
					}
					myHydrogens[index] = allHydrogens1[i];
					myHydIndex[index++] = i;
					allHydrogens1[i] = null;
				} else if ( neighbors[i].curPartner == 2 ) {
					allHydrogens1[i] = null;
				}
			}
		}
		if (index == 0) {
			double a, b, c, x, y, z, m;
			for ( int i = 0 ; i < 4 ; i++ ) {
				if (allHydrogens1[i] == null) {
					continue;
				}
				for ( int j = i+1 ; j < 4 ; j++ ) {
					if(allHydrogens1[j] == null) {
						continue;
					}
					// center around oxy
					a = allHydrogens1[i].coord[0] - coord[0];
					b = allHydrogens1[i].coord[1] - coord[1];
					c = allHydrogens1[i].coord[2] - coord[2];
					x = allHydrogens1[j].coord[0] - coord[0];
					y = allHydrogens1[j].coord[0] - coord[1];
					z = allHydrogens1[j].coord[0] - coord[2];
					x = -x-a;
					y = -y-b;
					z = -z-c;
					m = Math.sqrt((x*x)+(y*y)+(z*z));
					x = (x/m*MOL_DIPOLE) + dipole[0];
					y = (y/m*MOL_DIPOLE) + dipole[1];
					z = (z/m*MOL_DIPOLE) + dipole[2];
					m = Math.sqrt((x*x)+(y*y)+(z*z));
					if ( dipoles.containsKey(m) ) {
						m = m*(1.000000001+0.000000001*Math.random());
					}
					dipoles.put(m, new Double[] {(double) i, (double) j,m,x,y,z});
				}
			}
		} else if (index == 1) {
			double a, b, c, x, y, z, m;
			Hydrogen fixedHyd = myHydrogens[0];
			for ( int i = 0 ; i < 4 ; i++ ) {
				if (allHydrogens1[i] == null) {
					continue;
				}
				// center around oxy
				a = fixedHyd.coord[0] - coord[0];
				b = fixedHyd.coord[1] - coord[1];
				c = fixedHyd.coord[2] - coord[2];
				x = allHydrogens1[i].coord[0] - coord[0];
				y = allHydrogens1[i].coord[0] - coord[1];
				z = allHydrogens1[i].coord[0] - coord[2];
				x = -x-a;
				y = -y-b;
				z = -z-c;
				m = Math.sqrt((x*x)+(y*y)+(z*z));
				x = (x/m*MOL_DIPOLE) + dipole[0];
				y = (y/m*MOL_DIPOLE) + dipole[1];
				z = (z/m*MOL_DIPOLE) + dipole[2];
				m = Math.sqrt((x*x)+(y*y)+(z*z));
				if ( dipoles.containsKey(m) ) {
					m = m*(1.000000001+0.000000001*Math.random());
				}
				dipoles.put(m, new Double[] {(double) i, (double) myHydIndex[0],m,x,y,z});
			}
		} else {
			Hydrogen fixedHyd1 = myHydrogens[0];
			Hydrogen fixedHyd2 = myHydrogens[1];
			double a, b, c, x, y, z, m;
			a = fixedHyd1.coord[0] - coord[0];
			b = fixedHyd1.coord[1] - coord[1];
			c = fixedHyd1.coord[2] - coord[2];
			x = fixedHyd2.coord[0] - coord[0];
			y = fixedHyd2.coord[0] - coord[1];
			z = fixedHyd2.coord[0] - coord[2];
			x = -x-a;
			y = -y-b;
			z = -z-c;
			m = Math.sqrt((x*x)+(y*y)+(z*z));
			x = (x/m*MOL_DIPOLE) + dipole[0];
			y = (y/m*MOL_DIPOLE) + dipole[1];
			z = (z/m*MOL_DIPOLE) + dipole[2];
			m = Math.sqrt((x*x)+(y*y)+(z*z));
			if ( dipoles.containsKey(m) ) {
				m = m*(1.000000001+0.000000001*Math.random());
			}
			dipoles.put(m, new Double[] {(double) myHydIndex[0], (double) myHydIndex[1],m,x,y,z});
		}
		keys.addAll(dipoles.keySet());
		Collections.sort(keys);
		Comparator<Double> cmp = Collections.reverseOrder();
		Collections.sort(keys,cmp);
		if ( length(dipole) == 0.0 ) {
			Collections.swap(keys,0,1);
		}
		
	}

	private static double length(double[] dipole) {
		double x = dipole[0];
		double y = dipole[1];
		double z = dipole[2];
		return Math.sqrt((x*x)+(y*y)+(z*z));
	}
}
	

